'use strict';

/**
 * @author schukai GmbH
 */
import {assignToNamespace, Monster} from '../namespace.js';
import {isObject} from "../types/is.js";
import {AbstractConstraint} from "./abstract.js";

/**
 * Constraints are used to define conditions that must be met by the value of a variable.
 *
 * The uniform API of the constraints allows chains to be formed.
 *
 * You can call the method via the monster namespace `new Monster.Constraint.IsObject()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * console.log(new Monster.Constraint.IsObject())
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {IsObject} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/constraints/isobject.js';
 * console.log(new IsObject())
 * </script>
 * ```
 *
 * @example
 *
 * import {IsObject} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/constraints/isobject.js';
 *
 * new IsObject()
 * .isValid({})
 * .then(()=>console.log(true));
 * // ↦ true
 *
 *
 * new IsObject()
 * .isValid(99)
 * .catch(e=>console.log(e));
 * // ↦ 99
 *
 * @since 1.3.0
 * @copyright schukai GmbH
 * @memberOf Monster.Constraints
 * @summary A constraint to check if a value is an object
 */
class IsObject extends AbstractConstraint {

    /**
     * this method return a promise containing the result of the check.
     *
     * @param {*} value
     * @returns {Promise}
     */
    isValid(value) {
        if (isObject(value)) {
            return Promise.resolve(value);
        }

        return Promise.reject(value);
    }

}

assignToNamespace('Monster.Constraints', IsObject);
export {Monster, IsObject}
