'use strict';

/**
 * @author schukai GmbH
 */

import {internalSymbol} from "../constants.js";
import {assignToNamespace, Monster} from '../namespace.js';
import {Base} from "../types/base.js";
import {parseDataURL} from "../types/dataurl.js";
import {isString} from "../types/is.js";
import {ProxyObserver} from "../types/proxyobserver.js";
import {validateObject} from "../types/validate.js";
import {extend} from "./extend.js";
import {Pathfinder} from "./pathfinder.js";


/**
 * @private
 * @type {symbol}
 * @memberOf Monster.Data
 * @since 1.24.0
 */
const internalDataSymbol = Symbol('internalData');

/**
 * The datasource class is the basis for dealing with different data sources.
 * It provides a unified interface for accessing data
 *
 * You can create an object of this class using the monster namespace `Monster.Data.Datasource()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * Monster.Data.Datasource()
 * </script>
 * ```
 *
 * Alternatively you can import the class directly
 *
 * ```
 * <script type="module">
 * import {Datasource} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/data/datasource.js';
 * new Datasource()
 * </script>
 * ```
 *
 * @example
 *
 * import {Datasource} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/data/datasource.js'
 *
 * class MyDatasource extends Datasource {
 *  
 * }
 *
 * const ds = new MyDatasource();
 *
 * @since 1.22.0
 * @copyright schukai GmbH
 * @memberOf Monster.Data
 * @summary The datasource class encapsulates the access to data objects.
 */
class Datasource extends Base {

    /**
     *
     */
    constructor() {
        super();
        this[internalSymbol] = new ProxyObserver({
            'options': extend({}, this.defaults)
        });
        
        this[internalDataSymbol] = new ProxyObserver({

        });
        
    
    }

    /**
     * attach a new observer
     *
     * @param {Observer} observer
     * @returns {Datasource}
     */
    attachObserver(observer) {
        this[internalDataSymbol].attachObserver(observer)
        return this;
    }

    /**
     * detach a observer
     *
     * @param {Observer} observer
     * @returns {Datasource}
     */
    detachObserver(observer) {
        this[internalDataSymbol].detachObserver(observer)
        return this;
    }

    /**
     * @param {Observer} observer
     * @returns {boolean}
     */
    containsObserver(observer) {
        return this[internalDataSymbol].containsObserver(observer);
    }

    /**
     * Derived classes can override and extend this method as follows.
     *
     * ```
     * get defaults() {
     *    return Object.assign({}, super.defaults, {
     *        myValue:true
     *    });
     * }
     * ```
     */
    get defaults() {
        return {};
    }

    /**
     * Set option
     *
     * @param {string} path
     * @param {*} value
     * @return {Datasource}
     */
    setOption(path, value) {
        new Pathfinder(this[internalSymbol].getSubject()['options']).setVia(path, value);
        return this;
    }

    /**
     * @param {string|object} options
     * @return {Datasource}
     * @throws {Error} the options does not contain a valid json definition
     */
    setOptions(options) {

        if (isString(options)) {
            options = parseOptionsJSON(options)
        }

        const self = this;
        extend(self[internalSymbol].getSubject()['options'], self.defaults, options);

        return self;
    }

    /**
     * nested options can be specified by path `a.b.c`
     *
     * @param {string} path
     * @param {*} defaultValue
     * @return {*}
     */
    getOption(path, defaultValue) {
        let value;

        try {
            value = new Pathfinder(this[internalSymbol].getRealSubject()['options']).getVia(path);
        } catch (e) {

        }

        if (value === undefined) return defaultValue;
        return value;
    }

    /**
     * @throws {Error} this method must be implemented by derived classes.
     * @return {Promise}
     */
    read() {
        throw new Error("this method must be implemented by derived classes")
    }

    /**
     * @throws {Error} this method must be implemented by derived classes.
     * @return {Promise}
     */
    write() {
        throw new Error("this method must be implemented by derived classes")
    }


    /**
     * Returns real object
     * 
     * @return {Object|Array}
     */
    get() {
        const self = this;
        return self[internalDataSymbol].getRealSubject();
    }

    /**
     * @param {Object|Array} data
     * @return {Datasource}
     */
    set(data) {
        const self = this;
        self[internalDataSymbol].setSubject(data);
        return self;
    }

}

/**
 * @private
 * @param data
 * @return {Object}
 * @throws {Error} the options does not contain a valid json definition
 */
function parseOptionsJSON(data) {
    if (isString(data)) {

        // the configuration can be specified as a data url.
        try {
            let dataUrl = parseDataURL(data);
            data = dataUrl.content;
        } catch (e) {

        }


        try {
            let obj = JSON.parse(data);
            validateObject(obj);
            return obj;
        } catch (e) {
            throw new Error('the options does not contain a valid json definition (actual: ' + data + ').');
        }
    }

    return {};
}


assignToNamespace('Monster.Data', Datasource);
export {Monster, Datasource}
