'use strict';

/**
 * @author schukai GmbH
 */

import {internalSymbol} from "../../constants.js";
import {assignToNamespace, Monster} from '../../namespace.js';
import {validateString} from "../../types/validate.js";
import {Datasource} from "../datasource.js";

/**
 * @private
 * @type {symbol}
 */
const storageObjectSymbol = Symbol('storageObject');

/**
 * You can create an object of this class using the monster namespace `Monster.Data.Datasource.Storage()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * Monster.Data.Datasource.Storage()
 * </script>
 * ```
 *
 * Alternatively you can import the class directly
 *
 * ```
 * <script type="module">
 * import {Storage} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/data/datasource/storage.js';
 * new Storage()
 * </script>
 * ```
 *
 * @example
 *
 * import {Storage} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/data/datasource/storage.js';
 *
 * new Datasource();
 *
 * @since 1.22.0
 * @copyright schukai GmbH
 * @memberOf Monster.Data.Datasource
 * @summary The Storage class encapsulates the access to data objects over WebStorageAPI.
 */
class Storage extends Datasource {

    /**
     *
     * @param {string} key LocalStorage Key
     * @throws {TypeError} value is not a string
     */
    constructor(key) {
        super();
        this.setOption('key', validateString(key));
    }

    /**
     * @property {string} key=undefined LocalStorage Key
     */
    get defaults() {
        return Object.assign({}, super.defaults, {
            key: undefined,
        });
    }

    /**
     * @throws {Error} this method must be implemented by derived classes.
     * @return {external:Storage}
     * @private
     */
    [storageObjectSymbol]() {
        throw new Error("this method must be implemented by derived classes")
    }

    /**
     * @return {Promise}
     * @throws {Error} the options does not contain a valid json definition
     * @throws {TypeError} value is not a object
     * @throws {Error} the data cannot be read
     */
    read() {
        const self = this;

        const storage = self[storageObjectSymbol]();

        return new Promise(function (resolve) {
            const data = JSON.parse(storage.getItem(self.getOption('key')));
            self.set(data??{});
            resolve();
        })

    }

    /**
     * @return {Storage}
     * @throws {Error} the data cannot be written
     */
    write() {
        const self = this;

        const storage = self[storageObjectSymbol]();

        return new Promise(function (resolve) {

            const data = self.get();
            if (data === undefined) {
                storage.removeItem(self.getOption('key'));
            } else {
                storage.setItem(self.getOption('key'), JSON.stringify(data));
            }

            resolve();
        })
    }

    /**
     * @return {Storage}
     */
    getClone() {
        const self=this;
        return new Storage(self[internalSymbol].getRealSubject()['options'].key);
    } 
    
}


assignToNamespace('Monster.Data.Datasource', Storage);
export {Monster, Storage, storageObjectSymbol}
