'use strict';

/**
 * @author schukai GmbH
 */
import {assignToNamespace, Monster} from '../namespace.js';
import {isArray, isObject} from "../types/is.js";
import {typeOf} from "../types/typeof.js";

/**
 * Extend copies all enumerable own properties from one or
 * more source objects to a target object. It returns the modified target object.
 *
 * You can call the method via the monster namespace `Monster.Data.extend()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * Monster.Data.extend(a, b)
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {extend} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/data/extend.js';
 * extend(a, b)
 * </script>
 * ```
 *
 * @param {object} target
 * @param {object}
 * @return {object}
 * @since 1.10.0
 * @copyright schukai GmbH
 * @memberOf Monster.Data
 * @throws {Error} unsupported argument
 * @throws {Error} type mismatch
 */
function extend() {
    let o, i;

    for (i = 0; i < arguments.length; i++) {
        let a = arguments[i];

        if (!(isObject(a) || isArray(a))) {
            throw new Error('unsupported argument ' + JSON.stringify(a));
        }

        if (o === undefined) {
            o = a;
            continue;
        }

        for (let k in a) {

            let v = a?.[k];

            if (v === o?.[k]) {
                continue;
            }

            if ((isObject(v)&&typeOf(v)==='object') || isArray(v)) {

                if (o[k] === undefined) {
                    if (isArray(v)) {
                        o[k] = [];
                    } else {
                        o[k] = {};
                    }
                } else {
                    if (typeOf(o[k]) !== typeOf(v)) {
                        throw new Error("type mismatch: " + JSON.stringify(o[k]) + "(" + typeOf(o[k]) + ") != " + JSON.stringify(v) + "(" + typeOf(v) + ")");
                    }
                }

                o[k] = extend(o[k], v);

            } else {
                o[k] = v;
            }

        }
    }

    return o;
}


assignToNamespace('Monster.Data', extend);
export {Monster, extend}
