'use strict';

/**
 * @author schukai GmbH
 */
import {assignToNamespace, Monster} from '../namespace.js';
import {Base} from '../types/base.js';
import {isArray, isInteger, isObject, isPrimitive} from '../types/is.js';
import {Stack} from "../types/stack.js";
import {validateInteger, validateString} from '../types/validate.js';

/**
 * path separator
 *
 * @private
 * @type {string}
 */
export const DELIMITER = '.';

/**
 * @private
 * @type {string}
 */
export const WILDCARD = '*';

/**
 * You can call the method via the monster namespace `new Monster.Data.Pathfinder()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * console.log(new Monster.Data.Pathfinder())
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {Pathfinder} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/data/pathfinder.js';
 * console.log(new Pathfinder())
 * </script>
 * ```
 *
 * With the help of the pathfinder, values can be read and written from an object construct.
 *
 * ```
 * new Pathfinder({
 * a: {
 *     b: {
 *         f: [
 *             {
 *                 g: false,
 *             }
 *         ],
 *     }
 * }
 * }).getVia("a.b.f.0.g"); // ↦ false
 * ```
 *
 * if a value is not present or has the wrong type, a corresponding exception is thrown.
 *
 * ```
 * new Pathfinder({}).getVia("a.b.f.0.g"); // ↦ Error
 * ```
 *
 * The `Pathfinder.exists()` method can be used to check whether access to the path is possible.
 *
 * ```
 * new Pathfinder({}).exists("a.b.f.0.g"); // ↦ false
 * ```
 *
 * pathfinder can also be used to build object structures. to do this, the `Pathfinder.setVia()` method must be used.
 *
 * ```
 * obj = {};
 * new Pathfinder(obj).setVia('a.b.0.c', true); // ↦ {a:{b:[{c:true}]}}
 * ```
 *
 * @example
 *
 * import {Pathfinder} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/data/pathfinder.js';
 *
 * let value = new Pathfinder({
 * a: {
 *     b: {
 *         f: [
 *             {
 *                 g: false,
 *             }
 *         ],
 *     }
 * }
 * }).getVia("a.b.f.0.g");
 *
 *  console.log(value);
 *  // ↦ false
 *
 * try {
 *   new Pathfinder({}).getVia("a.b.f.0.g");  
 * } catch(e) {
 *   console.log(e.toString());
 *   // ↦ Error: the journey is not at its end (b.f.0.g)
 * }
 *
 * @example
 *
 * import {Pathfinder} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/data/pathfinder.js';
 *
 * let p = new Pathfinder({
 *                a: {
 *                    x: [
 *                        {c: 1}, {c: 2}
 *                    ],
 *                    y: true
 *                },
 *                b: {
 *                    x: [
 *                        {c: 1, d: false}, {c: 2}
 *                    ],
 *                    y: true
 *                },
 *            });
 *
 * let r = p.getVia("*.x.*.c");
 * console.log(r);
 *
 * @since 1.4.0
 * @copyright schukai GmbH
 * @memberOf Monster.Data
 */
class Pathfinder extends Base {

    /**
     * @param {array|object|Map|Set} value
     * @since 1.4.0
     * @throws  {Error} the parameter must not be a simple type
     **/
    constructor(object) {
        super();

        if (isPrimitive(object)) {
            throw new Error('the parameter must not be a simple type');
        }

        this.object = object;
        this.wildCard = WILDCARD;
    }

    /**
     * set wildcard
     *
     * @param {string} wildcard
     * @return {Pathfinder}
     * @since 1.7.0
     */
    setWildCard(wildcard) {
        validateString(wildcard);
        this.wildCard = wildcard;
        return this;
    }

    /**
     *
     * @param {string} path
     * @since 1.4.0
     * @returns {*}
     * @throws {TypeError} unsupported type
     * @throws {Error} the journey is not at its end
     * @throws {TypeError} value is not a string
     * @throws {TypeError} value is not an integer
     * @throws {Error} unsupported action for this data type
     */
    getVia(path) {
        return getValueViaPath.call(this, this.object, validateString(path));
    }

    /**
     *
     * @param {string} path
     * @param {*} value
     * @returns {Pathfinder}
     * @since 1.4.0
     * @throws {TypeError} unsupported type
     * @throws {TypeError} value is not a string
     * @throws {TypeError} value is not an integer
     * @throws {Error} unsupported action for this data type
     */
    setVia(path, value) {
        validateString(path);
        setValueViaPath.call(this, this.object, path, value);
        return this;
    }

    /**
     * Delete Via Path
     *
     * @param {string} path
     * @returns {Pathfinder}
     * @since 1.6.0
     * @throws {TypeError} unsupported type
     * @throws {TypeError} value is not a string
     * @throws {TypeError} value is not an integer
     * @throws {Error} unsupported action for this data type
     */
    deleteVia(path) {
        validateString(path);
        deleteValueViaPath.call(this, this.object, path);
        return this;
    }

    /**
     *
     * @param {string} path
     * @return {bool}
     * @throws {TypeError} unsupported type
     * @throws {TypeError} value is not a string
     * @throws {TypeError} value is not an integer
     * @since 1.4.0
     */
    exists(path) {
        validateString(path);
        try {
            getValueViaPath.call(this, this.object, path, true);
            return true;
        } catch (e) {

        }

        return false;
    }

}

assignToNamespace('Monster.Data', Pathfinder);
export {Monster, Pathfinder}

/**
 *
 * @param {*} subject
 * @param {string} path
 * @param {string} check
 * @return {Map}
 * @throws {TypeError} unsupported type
 * @throws {Error} the journey is not at its end
 * @throws {Error} unsupported action for this data type
 * @private
 */
function iterate(subject, path, check) {

    const result = new Map;

    if (isObject(subject) || isArray(subject)) {
        for (const [key, value] of Object.entries(subject)) {
            result.set(key, getValueViaPath.call(this, value, path, check))
        }
    } else {
        let key = path.split(DELIMITER).shift();
        result.set(key, getValueViaPath.call(this, subject, path, check));
    }

    return result;


}

/**
 *
 * @param {*} subject
 * @param [string} path
 * @param [boolean} check 
 * @returns {*}
 * @throws {TypeError} unsupported type
 * @throws {Error} the journey is not at its end
 * @throws {Error} unsupported action for this data type
 * @private
 */
function getValueViaPath(subject, path, check) {

    if (path === "") {
        return subject;
    }

    let parts = path.split(DELIMITER)
    let current = parts.shift();

    if (current === this.wildCard) {
        return iterate.call(this, subject, parts.join(DELIMITER), check);
    }

    if (isObject(subject) || isArray(subject)) {

        let anchor;
        if (subject instanceof Map || subject instanceof WeakMap) {
            anchor = subject.get(current);

        } else if (subject instanceof Set || subject instanceof WeakSet) {
            current = parseInt(current);
            validateInteger(current)
            anchor = [...subject]?.[current];

        } else if (typeof WeakRef === 'function' && subject instanceof WeakRef) {
            throw Error('unsupported action for this data type');

        } else if (isArray(subject)) {
            current = parseInt(current);
            validateInteger(current)
            anchor = subject?.[current];
        } else {
            anchor = subject?.[current];
        }

        if (isObject(anchor) || isArray(anchor)) {
            return getValueViaPath.call(this, anchor, parts.join(DELIMITER), check)
        }

        if (parts.length > 0) {
            throw Error("the journey is not at its end (" + parts.join(DELIMITER) + ")");
        }


        if (check === true) {
            const descriptor = Object.getOwnPropertyDescriptor(Object.getPrototypeOf(subject), current);

            if (!subject.hasOwnProperty(current) && descriptor === undefined) {
                throw Error('unknown value');
            }

        }

        return anchor;

    }

    throw TypeError("unsupported type " + typeof subject)

}

/**
 *
 * @param object
 * @param path
 * @param value
 * @returns {void}
 * @throws {TypeError} unsupported type
 * @throws {TypeError} unsupported type
 * @throws {Error} the journey is not at its end
 * @throws {Error} unsupported action for this data type
 * @private
 */
function setValueViaPath(object, path, value) {

    validateString(path);

    let parts = path.split(DELIMITER)
    let last = parts.pop();
    let subpath = parts.join(DELIMITER);

    let stack = new Stack()
    let current = subpath;
    while (true) {

        try {
            getValueViaPath.call(this, object, current, true)
            break;
        } catch (e) {

        }

        stack.push(current);
        parts.pop();
        current = parts.join(DELIMITER);

        if (current === "") break;
    }

    while (!stack.isEmpty()) {
        current = stack.pop();
        let obj = {};

        if (!stack.isEmpty()) {
            let n = stack.peek().split(DELIMITER).pop();
            if (isInteger(parseInt(n))) {
                obj = [];
            }

        }

        setValueViaPath.call(this, object, current, obj);
    }

    let anchor = getValueViaPath.call(this, object, subpath);

    if (!isObject(object) && !isArray(object)) {
        throw TypeError("unsupported type: " + typeof object);
    }

    if (anchor instanceof Map || anchor instanceof WeakMap) {
        anchor.set(last, value);
    } else if (anchor instanceof Set || anchor instanceof WeakSet) {
        anchor.append(value)

    } else if (typeof WeakRef === 'function' && anchor instanceof WeakRef) {
        throw Error('unsupported action for this data type');

    } else if (isArray(anchor)) {
        last = parseInt(last);
        validateInteger(last)
        assignProperty(anchor, last, value);
    } else {
        assignProperty(anchor, last, value);
    }


}

/**
 * @private
 * @param {object} object
 * @param {string} key
 * @param {*} value
 */
function assignProperty(object, key, value) {

    if (!object.hasOwnProperty(key)) {
        object[key] = value;
        return;
    }

    if (value === undefined) {
        delete object[key];
    }

    object[key] = value;

}

/**
 *
 * @param object
 * @param path
 * @returns {void}
 * @throws {TypeError} unsupported type
 * @throws {TypeError} unsupported type
 * @throws {Error} the journey is not at its end
 * @throws {Error} unsupported action for this data type
 * @since 1.6.0
 * @private
 */
function deleteValueViaPath(object, path) {

    const parts = path.split(DELIMITER)
    let last = parts.pop();
    const subpath = parts.join(DELIMITER);

    const anchor = getValueViaPath.call(this, object, subpath);

    if (anchor instanceof Map) {
        anchor.delete(last);
    } else if (anchor instanceof Set || anchor instanceof WeakMap || anchor instanceof WeakSet || (typeof WeakRef === 'function' && anchor instanceof WeakRef)) {
        throw Error('unsupported action for this data type');

    } else if (isArray(anchor)) {
        last = parseInt(last);
        validateInteger(last)
        delete anchor[last];
    } else {
        delete anchor[last];
    }


}
