'use strict';

/**
 * @author schukai GmbH
 */

import {assignToNamespace, Monster} from '../namespace.js';
import {Base} from '../types/base.js';
import {validateString} from '../types/validate.js';
import {Transformer} from './transformer.js';


const DELIMITER = '|';

/**
 * The pipe class makes it possible to combine several processing steps.
 *
 * You can call the method via the monster namespace `new Monster.Data.Pipe()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * new Monster.Data.Pipe()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {Pipe} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/data/pipe.js';
 * new Pipe()
 * </script>
 * ```
 *
 * A pipe consists of commands whose input and output are connected with the pipe symbol `|`.
 *
 * With the Pipe, processing steps can be combined. Here, the value of an object is accessed via the pathfinder (path command).
 * the word is then converted to uppercase letters and a prefix Hello is added. the two backslash safe the space char.
 *
 * @example
 * import {Pipe} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/data/pipe.js';
 *
 * let obj = {
 *    a: {
 *        b: {
 *            c: {
 *                d: "world"
 *            }
 *        }
 *    }
 * }
 *
 * console.log(new Pipe('path:a.b.c.d | toupper | prefix:Hello\\ ').run(obj));
 * // ↦ Hello WORLD
 *
 * @since 1.5.0
 * @copyright schukai GmbH
 * @memberOf Monster.Data
 */
class Pipe extends Base {

    /**
     *
     * @param {string} pipe a pipe consists of commands whose input and output are connected with the pipe symbol `|`.
     * @throws {TypeError}
     */
    constructor(pipe) {
        super();
        validateString(pipe);
        
        this.pipe = pipe.split(DELIMITER).map((v) => {
            return new Transformer(v);
        });


    }

    /**
     *
     * @param {string} name
     * @param {function} callback
     * @param {object} context
     * @returns {Transformer}
     * @throws {TypeError} value is not a string
     * @throws {TypeError} value is not a function
     */
    setCallback(name, callback, context) {

        for (const [, t] of Object.entries(this.pipe)) {
            t.setCallback(name, callback, context);
        }

        return this;
    }

    /**
     * run a pipe
     *
     * @param {*} value
     * @returns {*}
     */
    run(value) {
        return this.pipe.reduce((accumulator, transformer, currentIndex, array) => {
            return transformer.run(accumulator);
        }, value);
    }
}

assignToNamespace('Monster.Data', Pipe);
export {Monster, Pipe}
