'use strict';


/**
 * @author schukai GmbH
 */
import {assignToNamespace, Monster} from '../namespace.js';
import {isArray,isObject} from "../types/is.js";
import {validateInstance, validateString} from "../types/validate.js";
import {getDocument} from "./util.js";

/**
 * You can call the function via the monster namespace `new Monster.DOM.fireEvent()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * new Monster.DOM.fireEvent()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {fireEvent} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/dom/events.js';
 * fireEvent()
 * </script>
 * ```
 *
 * @param {HTMLElement|HTMLCollection|NodeList} element
 * @param {string} type
 * @return {void}
 * @since 1.10.0
 * @copyright schukai GmbH
 * @memberOf Monster.DOM
 * @throws {TypeError} value is not an instance of HTMLElement or HTMLCollection
 * @summary Construct and send and event
 */
function fireEvent(element, type) {

    const document = getDocument();

    if (element instanceof HTMLElement) {

        if (type === 'click') {
            element.click();
            return;
        }

        let event = new Event(validateString(type), {
            bubbles: true,
            cancelable: true,
        });

        element.dispatchEvent(event);

    } else if (element instanceof HTMLCollection || element instanceof NodeList) {
        for (let e of element) {
            fireEvent(e, type);
        }
    } else {
        throw new TypeError('value is not an instance of HTMLElement or HTMLCollection')
    }

}

/**
 * You can call the function via the monster namespace `new Monster.DOM.fireCustomEvent()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * new Monster.DOM.fireCustomEvent()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {fireCustomEvent} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/dom/events.js';
 * fireCustomEvent()
 * </script>
 * ```
 *
 * @param {HTMLElement|HTMLCollection|NodeList} element
 * @param {string} type
 * @return {void}
 * @since 1.29.0
 * @copyright schukai GmbH
 * @memberOf Monster.DOM
 * @throws {TypeError} value is not an instance of HTMLElement or HTMLCollection
 * @summary Construct and send and event
 */
function fireCustomEvent(element, type, detail) {

    const document = getDocument();

    if (element instanceof HTMLElement) {

        if (!isObject(detail)) {
            detail = {detail};
        }

        let event = new CustomEvent(validateString(type), {
            bubbles: true,
            cancelable: true,
            detail
        });

        element.dispatchEvent(event);

    } else if (element instanceof HTMLCollection || element instanceof NodeList) {
        for (let e of element) {
            fireCustomEvent(e, type, detail);
        }
    } else {
        throw new TypeError('value is not an instance of HTMLElement or HTMLCollection')
    }

}

/**
 * This function gets the path `Event.composedPath()` from an event and tries to find the next element
 * up the tree `element.closest()` with the attribute and value. If no value, or a value that is undefined or null,
 * is specified, only the attribute is searched.
 *
 * You can call the function via the monster namespace `new Monster.DOM.findTargetElementFromEvent()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * new Monster.DOM.findTargetElementFromEvent()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {findTargetElementFromEvent} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/dom/events.js';
 * findTargetElementFromEvent()
 * </script>
 * ```
 *
 * @since 1.14.0
 * @param {Event} event
 * @param {string} attributeName
 * @param {string|null|undefined} attributeValue
 * @throws {Error} unsupported event
 * @memberOf Monster.DOM
 * @throws {TypeError} value is not a string
 * @throws {TypeError} value is not an instance of HTMLElement
 * @summary Help function to find the appropriate control
 */
function findTargetElementFromEvent(event, attributeName, attributeValue) {
    validateInstance(event, Event);

    if (typeof event.composedPath !== 'function') {
        throw new Error('unsupported event');
    }

    const path = event.composedPath();

    // closest cannot be used here, because closest is not correct for slotted elements
    if (isArray(path)) {
        for (let i = 0; i < path.length; i++) {
            const o = path[i];

            if (o instanceof HTMLElement &&
                o.hasAttribute(attributeName)
                && (attributeValue === undefined || o.getAttribute(attributeName) === attributeValue)) {
                return o;
            }
        }
    }

    return undefined;

}


assignToNamespace('Monster.DOM', findTargetElementFromEvent, fireEvent, fireCustomEvent);
export {Monster, findTargetElementFromEvent, fireEvent, fireCustomEvent}
