'use strict';

/**
 * @author schukai GmbH
 */

import {parseLocale} from "../i18n/locale.js";
import {assignToNamespace, Monster} from '../namespace.js';
import {getDocument} from "./util.js";

/**
 * @private
 * @type {string}
 */
const DEFAULT_LANGUAGE = 'en';

/**
 * With this function you can read the language version set by the document.
 * For this the attribute `lang` in the html tag is read. If no attribute is set, `en` is used as default.
 * 
 * ```html
 * <html lang="en">
 * ```
 *
 * You can call the function via the monster namespace `new Monster.DOM.getLocaleOfDocument()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * new Monster.DOM.getLocaleOfDocument()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {getLocaleOfDocument} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/dom/locale.js';
 * new getLocaleOfDocument()
 * </script>
 * ```
 *
 * @since 1.13.0
 * @copyright schukai GmbH
 * @memberOf Monster.DOM
 * @throws {TypeError} value is not a string
 * @throws {Error} unsupported locale
 * @summary Tries to determine the locale used
 */
function getLocaleOfDocument() {

    const document = getDocument();

    let html = document.querySelector('html')
    if (html instanceof HTMLElement && html.hasAttribute('lang')) {
        let locale = html.getAttribute('lang');
        if (locale) {
            return new parseLocale(locale)
        }
    }

    return parseLocale(DEFAULT_LANGUAGE);
}

assignToNamespace('Monster.DOM', getLocaleOfDocument);
export {Monster, getLocaleOfDocument}
