'use strict';

/**
 * @author schukai GmbH
 */

import {internalStateSymbol} from "../../constants.js";
import {extend} from "../../data/extend.js";
import {assignToNamespace, Monster} from "../../namespace.js";
import {getGlobalFunction} from "../../types/global.js";
import {
    ATTRIBUTE_CLASS,
    ATTRIBUTE_ERRORMESSAGE,
    ATTRIBUTE_ID,
    ATTRIBUTE_SRC,
    ATTRIBUTE_TITLE,
    ATTRIBUTE_TYPE,
    TAG_SCRIPT
} from "../constants.js";
import {KEY_DOCUMENT, KEY_QUERY, referenceSymbol, Resource} from "../resource.js";


/**
 * This class is used by the resource manager to embed data.
 *
 * You can call the method via the monster namespace `new Monster.DOM.Resource.Data()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * new Monster.DOM.Resource.Data()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {Data} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/dom/resource/data.js';
 * new Data()
 * </script>
 * ```
 *
 * @since 1.25.0
 * @copyright schukai GmbH
 * @memberOf Monster.DOM.Resource
 * @summary A Data Resource class
 */
class Data extends Resource {

    /**
     * @property {string} mode=cors https://developer.mozilla.org/en-US/docs/Web/API/fetch
     * @property {string} credentials=same-origin https://developer.mozilla.org/en-US/docs/Web/API/fetch
     * @property {string} type=application/json {@link https://developer.mozilla.org/en-US/docs/Web/HTML/Element/script#attr-type}
     */
    get defaults() {
        return extend({}, super.defaults, {
            mode: 'cors',
            credentials: 'same-origin',
            type: 'application/json',
        })
    }

    /**
     *
     * @return {Monster.DOM.Resource.Data}
     */
    create() {
        createElement.call(this);
        return this;
    }

    /**
     * This method appends the HTMLElement to the specified document
     *
     * throws {Error} target not found
     * @return {Monster.DOM.Resource}
     */
    connect() {

        if (!(this[referenceSymbol] instanceof HTMLElement)) {
            this.create();
        }

        appendToDocument.call(this);
        return this;
    }

    /**
     * @return {string}
     */
    static getURLAttribute() {
        return ATTRIBUTE_SRC
    }

}

/**
 * @private
 * @return {Monster.DOM.Resource.Data}
 */
function createElement() {
    const self = this;

    const document = self.getOption(KEY_DOCUMENT);
    self[referenceSymbol] = document.createElement(TAG_SCRIPT);

    for (let key of [ATTRIBUTE_TYPE, ATTRIBUTE_ID, ATTRIBUTE_CLASS, ATTRIBUTE_TITLE]) {
        if (self.getOption(key) !== undefined) {
            self[referenceSymbol][key] = self.getOption(key);
        }
    }

    return self;
}


/**
 * @private
 * @return {Promise}
 * throws {Error} target not found
 */
function appendToDocument() {
    const self = this;

    const targetNode = document.querySelector(self.getOption(KEY_QUERY, 'head'))
    if (!(targetNode instanceof HTMLElement)) {
        throw new Error('target not found')
    }

    targetNode.appendChild(self[referenceSymbol]);

    getGlobalFunction('fetch')(self.getOption(ATTRIBUTE_SRC), {
        method: 'GET', // *GET, POST, PUT, DELETE, etc.
        mode: self.getOption('mode', 'cors'), // no-cors, *cors, same-origin
        cache: 'no-cache', // *default, no-cache, reload, force-cache, only-if-cached
        credentials: self.getOption('credentials', 'same-origin'), // include, *same-origin, omit
        headers: {
            'Accept': self.getOption('type', 'application/json')
        },
        redirect: 'follow', // manual, *follow, error
        referrerPolicy: 'no-referrer', // no-referrer, 
    }).then(response => {

        return response.text()


    }).then(text => {

        const textNode = document.createTextNode(text);
        self[referenceSymbol].appendChild(textNode);

        self[internalStateSymbol].getSubject()['loaded'] = true;


    }).catch(e => {
        self[internalStateSymbol].setSubject({
            loaded: true,
            error: e.toString(),
        })

        targetNode.setAttribute(ATTRIBUTE_ERRORMESSAGE, e.toString());
    })

    return self;
}


assignToNamespace('Monster.DOM.Resource', Data);
export {Monster, Data}
