'use strict';

/**
 * @author schukai GmbH
 */

import {extend} from "../../data/extend.js";
import {assignToNamespace, Monster} from "../../namespace.js";
import {
    ATTRIBUTE_CLASS,
    ATTRIBUTE_HREF,
    ATTRIBUTE_ID,
    ATTRIBUTE_NONCE, ATTRIBUTE_SRC,
    ATTRIBUTE_TITLE, ATTRIBUTE_TYPE,
    TAG_LINK
} from "../constants.js";
import {KEY_DOCUMENT, referenceSymbol, Resource} from "../resource.js";


/**
 * This class is used by the resource manager to embed external resources.
 *
 * You can call the method via the monster namespace `new Monster.DOM.Resource.Link()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * new Monster.DOM.Resource.Link()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {Link} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/dom/resource/link.js';
 * new Link()
 * </script>
 * ```
 *
 * @since 1.25.0
 * @copyright schukai GmbH
 * @memberOf Monster.DOM.Resource
 * @summary A Resource class
 * @see https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link
 */
class Link extends Resource {

    /**
     * @property {string} as {@link https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#attr-as}
     * @property {string} crossOrigin=anonymous {@link https://developer.mozilla.org/en-US/docs/Web/HTML/Element/script#attr-crossorigin}
     * @property {boolean} disabled
     * @property {string} href {@link https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#attr-href}
     * @property {string} hreflang {@link https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#attr-hreflang}
     * @property {string} imagesizes {@link https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#attr-imagesizes}
     * @property {string} imagesrcset {@link https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#attr-imagesrcset}
     * @property {string} integrity {@link https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#attr-integrity}
     * @property {string} media {@link https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#attr-media}
     * @property {string} prefetch {@link https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#attr-prefetch}
     * @property {string} referrerpolicy {@link https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#attr-referrerpolicy}
     * @property {string} rel {@link https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#attr-rel}
     * @property {string} type {@link https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#attr-type}
     * @property {string} sizes {@link https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link#attr-sizes}
     * @property {string} nonce {@link https://developer.mozilla.org/en-US/docs/Web/HTML/Element/script#attr-nonce}
     */
    get defaults() {
        return extend({}, super.defaults, {
            as: undefined,
            crossOrigin: 'anonymous',
            disabled: undefined,
            href: undefined,
            hreflang: undefined,
            imagesizes: undefined,
            imagesrcset: undefined,
            integrity: undefined,
            media: undefined,
            prefetch: undefined,
            referrerpolicy: undefined,
            rel: undefined,
            sizes: undefined,
            type: undefined,
            nonce: undefined
        })
    }

    /**
     *
     * @return {Monster.DOM.Resource.Link}
     */
    create() {
        createElement.call(this);
        return this;
    }

    /**
     * @return {string}
     */
    static getURLAttribute() {
        return ATTRIBUTE_HREF
    }

}

/**
 * @private
 * @return {Monster.DOM.Resource.Link}
 */
function createElement() {
    const self = this;

    const document = self.getOption(KEY_DOCUMENT);
    self[referenceSymbol] = document.createElement(TAG_LINK);

    for (let key of ['as','crossOrigin','disabled','href','hreflang','imagesizes','imagesrcset','integrity','media','prefetch','referrerpolicy','sizes','rel','type',ATTRIBUTE_HREF,ATTRIBUTE_ID,ATTRIBUTE_CLASS,ATTRIBUTE_TITLE,ATTRIBUTE_NONCE]) {
        if (self.getOption(key) !== undefined) {
            self[referenceSymbol][key] = self.getOption(key);
        }
    }    
    
    return self;
}


assignToNamespace('Monster.DOM.Resource', Link);
export {Monster, Link}
