'use strict';

/**
 * @author schukai GmbH
 */

import {extend} from "../data/extend.js";
import {assignToNamespace, Monster} from "../namespace.js";
import {BaseWithOptions} from "../types/basewithoptions.js";
import {getGlobalObject} from "../types/global.js";
import {isArray} from "../types/is.js";
import {ATTRIBUTE_HREF, ATTRIBUTE_SRC} from "./constants.js";
import {Resource} from "./resource.js";
import {Data} from "./resource/data.js";
import {Stylesheet} from "./resource/link/stylesheet.js";
import {Script} from "./resource/script.js";


/**
 * You can call the method via the monster namespace `new Monster.DOM.ResourceManager()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * new Monster.DOM.ResourceManager()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {Resource} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/dom/resourcemanager.js';
 * new ResourceManager()
 * </script>
 * ```
 *
 * @since 1.25.0
 * @copyright schukai GmbH
 * @memberOf Monster.DOM
 * @summary A Resource class
 */
class ResourceManager extends BaseWithOptions {

    /**
     *
     * @param {Object} options
     * throw {Error} unsupported document type
     */
    constructor(options) {
        super(options);

        if (!(this.getOption('document') instanceof Document)) {
            throw new Error('unsupported document type')
        }


    }

    /**
     * @property {string} baseurl
     */
    getBaseURL() {
        this.getOption('document')?.baseURL;
    }

    /**
     *
     * @property {HTMLDocument} document=document Document
     * @property {Object} resources
     * @property {Array} resources.scripts=[] array with {@link Monster.DOM.Resource.Script} objects
     * @property {Array} resources.stylesheets=[] array with {@link Monster.DOM.Resource.Link.Stylesheet} objects
     * @property {Array} resources.data=[] array with {@link Monster.DOM.Resource.Data} objects
     */
    get defaults() {
        return Object.assign({}, super.defaults, {
            document: getGlobalObject('document'),
            resources: {
                scripts: [],
                stylesheets: [],
                data: []
            }
        })
    }

    /**
     * Append Tags to DOM
     *
     * @return {Monster.DOM.ResourceManager}
     * @throws {Error} unsupported resource definition
     */
    connect() {
        runResourceMethod.call(this, 'connect');
        return this;
    }

    /**
     * Check if available
     *
     * @return {Promise}
     * @throws {Error} unsupported resource definition
     */
    available() {
        return Promise.all(runResourceMethod.call(this, 'available'));
    }

    /**
     * Add a script
     *
     * @param {string|URL} url
     * @param [Object|undefined} options
     * @return {Monster.DOM.ResourceManager}
     * @see Monster.DOM.Resource.Script
     */
    addScript(url, options) {
        return addResource.call(this, 'scripts', url, options);
    }


    /**
     * Add Stylesheet
     *
     * @param {string|URL} url
     * @param [Object|undefined} options
     * @return {Monster.DOM.ResourceManager}
     * @see Monster.DOM.Resource.Link.Stylesheet
     */
    addStylesheet(url, options) {
        return addResource.call(this, 'stylesheets', url, options);
    }

    /**
     * Add Data Tag
     *
     * @param {string|URL} url
     * @param [Object|undefined} options
     * @return {Monster.DOM.ResourceManager}
     * @see Monster.DOM.Resource.Data
     */
    addData(url, options) {
        return addResource.call(this, 'data', url, options);
    }


}

/**
 * @private
 * @param {string} method
 * @return {Array}
 */
function runResourceMethod(method) {
    const self = this;

    const result = [];

    for (const type of ['scripts', 'stylesheets', 'data']) {
        const resources = self.getOption('resources.' + type);
        if (!isArray(resources)) {
            continue;
        }

        for (const resource of resources) {
            if (!(resource instanceof Resource)) {
                throw new Error('unsupported resource definition')
            }

            result.push(resource[method]());
        }

    }

    return result;
}

/**
 *
 * @param {string} type
 * @param {string|URL} url
 * @param [Object|undefined} options
 * @return {Monster.DOM.ResourceManager}
 * @private
 */
function addResource(type, url, options) {
    const self = this;

    if (url instanceof URL) {
        url = url.toString();
    }

    options = options || {}

    let resource;
    switch (type) {
        case 'scripts':
            resource = new Script(extend({}, options, {[ATTRIBUTE_SRC]: url}))
            break;
        case 'stylesheets':
            resource = new Stylesheet(extend({}, options, {[ATTRIBUTE_HREF]: url}))
            break;
        case 'data':
            resource = new Data(extend({}, options, {[ATTRIBUTE_SRC]: url}))
            break;
        default:
            throw new Error('unsupported type ' + type)
    }

    (self.getOption('resources')?.[type]).push(resource);
    return self;
}


assignToNamespace('Monster.DOM', ResourceManager);
export {Monster, ResourceManager}
