'use strict';

/**
 * @author schukai GmbH
 */
import {assignToNamespace, Monster} from '../namespace.js';
import {Base} from '../types/base.js';
import {getGlobalObject} from '../types/global.js';
import {validateString} from "../types/validate.js";
import {ATTRIBUTE_THEME_NAME, DEFAULT_THEME} from "./constants.js";


/**
 * You can call the method via the monster namespace `new Monster.DOM.Theme()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * console.log(new Monster.DOM.Theme())
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {Theme} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/dom/theme.js';
 * console.log(new Theme())
 * </script>
 * ```
 *
 * @example
 *
 * import {getDocumentTheme} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/dom/theme.js';
 *
 * const theme = getDocumentTheme();
 * console.log(theme.getName());
 * // ↦ monster
 *
 * @since 1.7.0
 * @copyright schukai GmbH
 * @memberOf Monster.DOM
 * @summary A theme class
 */
class Theme extends Base {

    /**
     *
     * @param name
     * @throws {TypeError} value is not a string
     */
    constructor(name) {
        super();
        validateString(name);
        this.name = name;
    }

    /**
     *
     * @returns {string}
     */
    getName() {
        return this.name;
    }

}

/**
 * The theming used in the document can be defined via the html-tag.
 * The theming is specified via the attribute `data-monster-theme-name`.
 *
 * As name for a theme all characters are valid, which are also allowed for a HTMLElement-ID.
 *
 * ```
 * <html data-monster-theme-name="my-theme">
 * ```
 *
 * the default theme name is `monster`.
 *
 * @return {Theme}
 * @memberOf Monster.DOM
 * @since 1.7.0
 */
function getDocumentTheme() {
    let document = getGlobalObject('document');
    let name = DEFAULT_THEME;

    let element = document.querySelector('html');
    if (element instanceof HTMLElement) {
        let theme = element.getAttribute(ATTRIBUTE_THEME_NAME);
        if (theme) {
            name = theme;
        }
    }

    return new Theme(name);

}

assignToNamespace('Monster.DOM', Theme, getDocumentTheme);
export {Monster, Theme, getDocumentTheme}
