'use strict';

/**
 * @author schukai GmbH
 */


import {internalSymbol} from "../constants.js";
import {extend} from "../data/extend.js";
import {assignToNamespace, Monster} from '../namespace.js';
import {Formatter as TextFormatter} from "../text/formatter.js";
import {validateInstance, validateString} from "../types/validate.js";
import {Translations} from "./translations.js";

/**
 * @private
 * @type {symbol}
 */
const internalTranslationSymbol = Symbol('internalTranslation')

/**
 * The Formatter extends the Text.Formatter with the possibility to replace the key by a translation.
 *
 * You can create an instance via the monster namespace `new Monster.I18n.Formatter()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * new Monster.I18n.Formatter()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {Formatter} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/i18n/formatter.js';
 * new Formatter()
 * </script>
 * ```
 *
 * @example
 *
 * import {Formatter} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/i18n/formatter.js';
 * import {Translations} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/i18n/translations.js';
 *
 * const translations = new Translations('en')
 *     .assignTranslations({
 *         thekey: "${animal} has eaten the ${food}!"
 *     });
 *
 * new Formatter({}, translations).format("thekey:animal=dog::food=cake")
 * // ↦ dog has eaten the cake!
 *
 * @since 1.26.0
 * @copyright schukai GmbH
 * @memberOf Monster.I18n
 */
class Formatter extends TextFormatter {

    /**
     * Default values for the markers are `${` and `}`
     *
     * @param {object} object
     * @throws {TypeError} value is not a object
     */
    constructor(object, translation, options) {
        super(object, options);
        this[internalTranslationSymbol] = validateInstance(translation, Translations);
    }

    /**
     * @property {object} marker
     * @property {array} marker.open=["i18n{","${"]
     * @property {array} marker.close=["${"]
     * @property {object} parameter
     * @property {string} parameter.delimiter="::"
     * @property {string} parameter.assignment="="
     * @property {object} callbacks
     * @property {function} callbacks.i18n=()=>{}
     */
    get defaults() {
        const self = this;
        return extend({}, super.defaults, {
            callbacks: {
                i18n: (value) => {
                    return self[internalTranslationSymbol].getText(validateString(value));
                }
            },
            marker: {
                open: ['i18n{', '${'],
                close: ['}'],
            },
        })
    }

    /**
     *
     * @param {string} text
     * @return {string}
     * @throws {TypeError} value is not a string
     * @throws {Error} too deep nesting
     * @throws {Error} key not found
     * @throws {Error} the closing marker is missing
     */
    format(text) {
        validateString(text)

        const openMarker = this[internalSymbol]['marker']['open']?.[0];
        const closeMarker = this[internalSymbol]['marker']['close']?.[0];

        if (text.indexOf(openMarker) === 0) {
            text = text.substring(openMarker.length);

            if (text.indexOf(closeMarker) === text.length - closeMarker.length) {
                text = text.substring(0, text.length - closeMarker.length);
            } else {
                throw new Error("the closing marker is missing")
            }
        }


        const parts = validateString(text).split('::')
        const translationKey = parts.shift().trim(); // key value delimiter
        const parameter = parts.join('::').trim();


        let assembledText = openMarker + 'static:' + translationKey + ' | call:i18n';
        if (parameter.length > 0) {
            assembledText += '::' + parameter;
        }
        assembledText += closeMarker;
        return super.format(assembledText);
    }


}

assignToNamespace('Monster.I18n', Formatter);
export {Monster, Formatter}
