'use strict';

/**
 * @author schukai GmbH
 */

import {assignToNamespace, Monster} from '../../namespace.js';
import {Base} from '../../types/base.js';
import {getGlobalObject} from "../../types/global.js";
import {Handler} from '../handler.js';
import {LogEntry} from "../logentry.js";

/**
 * You can create an object of the class simply by using the namespace `new Monster.Logging.Handler.ConsoleHandler()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * console.log(new Monster.Logging.Handler.ConsoleHandler())
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this class individually.
 *
 * ```
 * <script type="module">
 * import {ConsoleHandler} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/logging/handler/console.js';
 * console.log(new ConsoleHandler())
 * </script>
 * ```
 *
 * @since 1.5.0
 * @copyright schukai GmbH
 * @memberOf Monster.Logging.Handler
 */
class ConsoleHandler extends Handler {
    constructor() {
        super();
    }


    /**
     * This is the central log function. this method must be
     * overwritten by derived handlers with their own logic.
     *
     * ALL > TRACE > DEBUG > INFO > WARN > ERROR > FATAL > OFF  (ALL = 0xff;OFF = 0x00;
     *
     * @param {LogEntry} entry
     * @returns {boolean}
     */
    log(entry) {
        if (super.log(entry)) {
            let console = getGlobalObject('console');
            if (!console) return false;
            console.log(entry.toString());
            return true;
        }

        return false;
    }

}


assignToNamespace('Monster.Logging.Handler', ConsoleHandler);
export {Monster, ConsoleHandler};
