'use strict';

/**
 * @author schukai GmbH
 */

import {internalSymbol} from "../constants.js";
import {extend} from "../data/extend.js";
import {Pathfinder} from "../data/pathfinder.js";
import {assignToNamespace, Monster} from '../namespace.js';
import {Base} from "./base.js";
import {validateObject} from "./validate.js";

/**
 * This is the base class with options from which some monster classes are derived.
 *
 * This class is actually only used as a base class.
 * 
 * However, you can also create an instance directly via the monster namespace `new Monster.Types.BaseWithOptions()`.
 *
 * ```html
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * new Monster.Types.BaseWithOptions()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```html
 * <script type="module">
 * import {BaseWithOptions} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/basewithoptions.js';
 * new BaseWithOptions()
 * </script>
 * ```
 * 
 * Classes that require the possibility of options can be derived directly from this class.
 * Derived classes almost always override the `defaul` getter with their own values.
 *
 * ```javascript
 * class My extends BaseWithOptions {
 *    get defaults() {
 *        return Object.assign({}, super.defaults, {
 *            mykey: true
 *        });
 *    }  
 * }
 * ```
 * 
 * The class was formerly called Object.
 *
 * @since 1.13.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types
 */
class BaseWithOptions extends Base {

    /**
     *
     * @param {object} options
     */
    constructor(options) {
        super();

        if (options === undefined) {
            options = {};
        }

        this[internalSymbol] = extend({}, this.defaults, validateObject(options));

    }

    /**
     * This getter provides the options. Derived classes overwrite 
     * this getter with their own values. It is good karma to always include 
     * the values from the parent class.
     * 
     * ```javascript
     * get defaults() {
     *     return Object.assign({}, super.defaults, {
     *         mykey: true
     *     });
     * }
     * 
     * ```
     *
     * @return {object}
     */
    get defaults() {
        return {}
    }

    /**
     * nested options can be specified by path `a.b.c`
     *
     * @param {string} path
     * @param {*} defaultValue
     * @return {*}
     * @since 1.10.0
     */
    getOption(path, defaultValue) {
        let value;

        try {
            value = new Pathfinder(this[internalSymbol]).getVia(path);
        } catch (e) {

        }

        if (value === undefined) return defaultValue;
        return value;
    }


}

assignToNamespace('Monster.Types', BaseWithOptions);
export {Monster, BaseWithOptions}
