'use strict';

import {assignToNamespace} from "../namespace.js";


/**
 * @author schukai GmbH
 */
import {Monster, validateString} from "./validate.js";

/**
 * You can call the function via the monster namespace `Monster.Types.toBinary()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * Monster.Types.toBinary()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {toBinary} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/binary.js';
 * toBinary()
 * </script>
 * ```
 *
 * @param {String} binary
 * @return {String}
 * @throws {TypeError} value is not a string
 * @memberOf Monster.Types
 * @since 1.18.0
 */
function toBinary(string) {
    const codeUnits = new Uint16Array(validateString(string).length);
    for (let i = 0; i < codeUnits.length; i++) {
        codeUnits[i] = string.charCodeAt(i);
    }

    const charCodes = new Uint8Array(codeUnits.buffer);
    let result = '';

    for (let i = 0; i < charCodes.byteLength; i++) {
        result += String.fromCharCode(charCodes[i]);
    }

    return result;
}

/**
 * You can call the function via the monster namespace `Monster.Types.fromBinary()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * Monster.Types.fromBinary()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {fromBinary} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/binary.js';
 * fromBinary()
 * </script>
 * ```
 *
 * @param {String} binary
 * @return {String}
 * @throws {TypeError} value is not a string
 * @memberOf Monster.Types
 * @since 1.18.0
 */
function fromBinary(binary) {
    const bytes = new Uint8Array(validateString(binary).length);
    for (let i = 0; i < bytes.length; i++) {
        bytes[i] = binary.charCodeAt(i);
    }
    const charCodes = new Uint16Array(bytes.buffer);
    let result = '';
    for (let i = 0; i < charCodes.length; i++) {
        result += String.fromCharCode(charCodes[i]);
    }
    return result;
}


assignToNamespace('Monster.Types', toBinary, fromBinary);
export {Monster, toBinary, fromBinary};
