'use strict';

import {assignToNamespace} from "../namespace.js";
/**
 * @author schukai GmbH
 */
import {Base, Monster} from "./base.js";
import {isString} from "./is.js";
import {MediaType, parseMediaType} from "./mediatype.js";
import {validateBoolean, validateInstance, validateString} from "./validate.js";

/**
 * @private
 * @type {symbol}
 */
const internal = Symbol('internal');

/**
 * You can create an object via the monster namespace `new Monster.Types.DataUrl()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * new Monster.Types.DataUrl()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {DataUrl} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/dataurl.js';
 * new DataUrl()
 * </script>
 * ```
 *
 * @since 1.8.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types
 * @see https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/Data_URIs
 * @see https://datatracker.ietf.org/doc/html/rfc2397
 */
class DataUrl extends Base {

    /**
     *
     * @param {String} content
     * @param {String|Monster.Types.MediaType} mediatype
     * @param {boolean} base64=true
     */
    constructor(content, mediatype, base64) {
        super();

        if (isString(mediatype)) {
            mediatype = parseMediaType(mediatype);
        }

        this[internal] = {
            content: validateString(content),
            mediatype: validateInstance(mediatype, MediaType),
            base64: validateBoolean(base64 === undefined ? true : base64)
        }


    }

    get content() {
        return this[internal].base64 ? atob(this[internal].content) : this[internal].content;
    }

    get mediatype() {
        return this[internal].mediatype;
    }


    /**
     *
     * @return {string}
     * @see https://datatracker.ietf.org/doc/html/rfc2397
     */
    toString() {

        let content = this[internal].content;

        if (this[internal].base64 === true) {
            content = ';base64,' + content;
        } else {
            content = ',' + encodeURIComponent(content);
        }

        return 'data:' + this[internal].mediatype.toString() + content;
    }

}

/**
 * You can call the function via the monster namespace `Monster.Types.parseDataURL()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * console.log(Monster.Types.parseDataURL())
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {parseDataURL} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/dataurl.js';
 * console.log(parseDataURL())
 * </script>
 * ```
 *
 * Specification:
 *
 * ```
 * dataurl    := "data:" [ mediatype ] [ ";base64" ] "," data
 * mediatype  := [ type "/" subtype ] *( ";" parameter )
 * data       := *urlchar
 * parameter  := attribute "=" value
 * ```
 *
 * @param {String} dataurl
 * @return {Monster.Types.DataUrl}
 * @see https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/Data_URIs
 * @see https://datatracker.ietf.org/doc/html/rfc2397
 * @throws {TypeError} incorrect or missing data protocol
 * @throws {TypeError} malformed data url
 * @memberOf Monster.Types
 */
function parseDataURL(dataurl) {

    validateString(dataurl);

    dataurl = dataurl.trim();

    if (dataurl.substring(0, 5) !== 'data:') {
        throw new TypeError('incorrect or missing data protocol')
    }

    dataurl = dataurl.substring(5);

    let p = dataurl.indexOf(',');
    if (p === -1) {
        throw new TypeError('malformed data url')
    }

    let content = dataurl.substring(p + 1);
    let mediatypeAndBase64 = dataurl.substring(0, p).trim();
    let mediatype = 'text/plain;charset=US-ASCII';
    let base64Flag = false;

    if (mediatypeAndBase64 !== "") {
        mediatype = mediatypeAndBase64;
        if (mediatypeAndBase64.endsWith('base64')) {
            let i = mediatypeAndBase64.lastIndexOf(';');
            mediatype = mediatypeAndBase64.substring(0, i);
            base64Flag = true;
        } else {
            content = decodeURIComponent(content);
        }

        mediatype = parseMediaType(mediatype);
    } else {
        content = decodeURIComponent(content);
    }

    return new DataUrl(content, mediatype, base64Flag);


}


assignToNamespace('Monster.Types', parseDataURL, DataUrl);
export {Monster, parseDataURL, DataUrl};
