'use strict';

/**
 * @author schukai GmbH
 */


import {assignToNamespace, Monster} from '../namespace.js';
import {Base} from './base.js';
import {isPrimitive} from "./is.js";
import {NodeList} from './nodelist.js';
import {validateInstance} from './validate.js';

/**
 * @private
 * @type {symbol}
 */
const internalValueSymbol = Symbol('internalData');

/**
 * @private
 * @type {symbol}
 */
const treeStructureSymbol = Symbol('treeStructure');


/**
 * You can create the instance via the monster namespace `new Monster.Types.Node()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * new Monster.Types.Node()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {Node} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/node.js';
 * new Node()
 * </script>
 * ```
 *
 * @since 1.26.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types
 * @summary A Node Class
 * @see https://developer.mozilla.org/de/docs/Web/JavaScript/Reference/Iteration_protocols
 */
class Node extends Base {

    /**
     * @param {*} [value]
     */
    constructor(value) {
        super();
        this[internalValueSymbol] = value;

        this[treeStructureSymbol] = {
            parent: null,
            childNodes: new NodeList,
            level: 0
        }

    }

    /**
     * @property {*}
     */
    get value() {
        return this[internalValueSymbol];
    }

    /**
     * @property {*}
     */
    set value(value) {
        this[internalValueSymbol] = value;
    }

    /**
     * @property {Monster.Types.Node|null}
     */
    get parent() {
        return this[treeStructureSymbol].parent;
    }

    /**
     * @property {integer}
     */
    get level() {
        return this[treeStructureSymbol].level;
    }

    /**
     *
     * @property {NodeList}
     */
    get childNodes() {
        return this[treeStructureSymbol].childNodes;
    }

    /**
     *
     * @property {NodeList}
     */
    set childNodes(childNodes) {
        this[treeStructureSymbol].childNodes = validateInstance(childNodes, NodeList);
        setChildLevelAndParent.call(this, this, 1);
    }

    /**
     * @return {Monster.Types.Node}
     * @param {Node} node
     */
    appendChild(node) {
        this[treeStructureSymbol].childNodes.add(validateInstance(node, Node));
        node[treeStructureSymbol].parent = this;

        node[treeStructureSymbol].level = this.level + 1;
        setChildLevelAndParent.call(this, node, 1);
        return this;
    }

    /**
     * @return {Monster.Types.Node}
     * @param {Node} node
     */
    removeChild(node) {
        this[treeStructureSymbol].childNodes.remove(validateInstance(node, Node));
        node[treeStructureSymbol].parent = null;

        node[treeStructureSymbol].level = 0;
        setChildLevelAndParent.call(this, node, -1);
        return this;
    }

    /**
     *
     * @return {boolean}
     */
    hasChildNodes() {
        return this[treeStructureSymbol].childNodes.length > 0;
    }

    /**
     * @return {Monster.Types.Node}
     * @param {Node} node
     */
    hasChild(node) {
        return this[treeStructureSymbol].childNodes.has(validateInstance(node, Node));
    }

    /**
     * @since 1.28.0
     * @return {string}
     */
    toString() {

        let parts = [];
        if (this[internalValueSymbol]) {
            let label = this[internalValueSymbol];
            if (!isPrimitive(label)) label = JSON.stringify(this[internalValueSymbol])

            parts.push( label);
        }

        if (!this.hasChildNodes()) {
            return parts.join("\n");
        }

        let count = this.childNodes.length,
            counter = 0;

        for (const node of this.childNodes) {
            counter++;
            const prefix = (count === counter ? '└' : '├').padStart(2 * node.level, ' |');
            parts.push(prefix + node.toString());
        }

        return parts.join("\n");
    }

}

/**
 * @private
 * @param {Node} node
 * @param {int} operand
 * @return {setChildLevelAndParent}
 */
function setChildLevelAndParent(node, operand) {
    const self = this;

    if (node !== this) {
        node[treeStructureSymbol].parent = this
    }

    node[treeStructureSymbol].childNodes.forEach(function (child) {
        child[treeStructureSymbol].parent = node;
        child[treeStructureSymbol].level = node[treeStructureSymbol].level + operand;
        setChildLevelAndParent.call(self, child, operand);
    });
    return this;
}

assignToNamespace('Monster.Types', Node);
export {Monster, Node}
