'use strict';

/**
 * @author schukai GmbH
 */
import {assignToNamespace, Monster} from '../namespace.js';
import {isArray, isInstance} from "./is.js";
import {Node} from "./node.js";
import {validateInstance} from "./validate.js";

/**
 * You can create the instance via the monster namespace `new Monster.Types.NodeList()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * new Monster.Types.NodeList()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {NodeList} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/nodelist.js';
 * new NodeList()
 * </script>
 * ```
 *
 * @since 1.26.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types
 * @summary A NodeList class
 */
class NodeList extends Set {

    /**
     * @throws {Error} invalid value type
     * @param {NodeList|Node|Array<Node>}values
     */
    constructor(values) {
        super();

        const self = this

        if(values===undefined) return;
        
        if (isArray(values)) {
            values.forEach(value => self.add(value));
        } else if (isInstance(values, NodeList)) {
            values.forEach(value => self.add(value));
        } else if (isInstance(values, Node)) {
            self.add(values);
        } else {
            throw new Error('invalid value type');
        }
    }

    /**
     *
     * @param {Node} node
     * @return {Monster.Types.NodeList}
     */
    add(node) {
        super.add(validateInstance(node, Node));
        return this;
    }

    /**
     * @param {Node} node
     * @returns {NodeList}
     */
    remove(node) {
        super.delete(validateInstance(node, Node));
        return this;
    }

    /**
     * @param {Node} node
     * @returns {boolean}
     */
    has(node) {
        return super.has(validateInstance(node, Node));
        return false;
    }

    /**
     * @returns {NodeList}
     */
    clear() {
        super.clear();
        return this;
    }

    /**
     * @returns {NodeList}
     */
    toArray() {
        return Array.from(this);
    }

    /**
     * @returns {NodeList}
     */
    toJSON() {
        return this.toArray();
    }

    /**
     * @returns {NodeList}
     */
    toString() {
        let parts = [];
        
        for(const node of this.toArray()) {
            parts.push(node.toString()) 
        }
        
        return parts.join("\n");
    }

    get length() {
        return super.size;
    }
}


assignToNamespace('Monster.Types', NodeList);
export {Monster, NodeList}
