'use strict';

/**
 * @author schukai GmbH
 */


import {internalSymbol} from "../constants.js";
import {assignToNamespace, Monster} from '../namespace.js';
import {Base} from './base.js';
import {isInstance} from "./is.js";
import {Node} from "./node.js";
import {NodeList} from "./nodelist.js";
import {validateInstance} from "./validate.js";


/**
 * @private
 * @type {symbol}
 */
const isNodeListSymbol = Symbol('isNodeList');

/**
 * You can create the instance via the monster namespace `new Monster.Types.NodeRecursiveIterator()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * new Monster.Types.NodeRecursiveIterator()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {NodeRecursiveIterator} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/noderecursiveiterator.js';
 * new NodeRecursiveIterator()
 * </script>
 * ```
 *
 * @example
 * import {Node} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/node.js';
 * import {NodeRecursiveIterator} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/noderecursiveiterator.js';
 *
 * const node =new Node('1')
 *
 * // 1
 * // 2
 * // ├ 2.1
 * // ├ 2.2
 * // └ 2.3
 * // 3
 * // 4
 * // ├ 4.1
 * // └ 4.2
 * node.appendChild(
 * (new Node('2'))
 *       .appendChild(new Node('2.1'))
 *       .appendChild(new Node('2.2'))
 *       .appendChild(new Node('2.3')))
 * .appendChild(new Node('3'))
 * .appendChild(new Node('4')
 *       .appendChild(new Node('4.1'))
 *       .appendChild(new Node('4.2')));
 *
 * const iterator = new NodeRecursiveIterator(node);
 *
 * const result = [];
 * for (const n of iterator) {
 *      result.push(n.value);
 * }
 *
 * // ↦ ['1', '2', '2.1', '2.2', '2.3', '3', '4', '4.1', '4.2']
 *
 * @since 1.26.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types
 * @summary An iterator to run recursively through a tree of nodes
 */
class NodeRecursiveIterator extends Base {

    /**
     * @param {Node} [data]
     */
    constructor(node) {
        super();

        this[isNodeListSymbol] = false;

        // iterator is a nodelist
        if (isInstance(node, NodeList)) {
            let children = node;
            node = new Node();
            node.childNodes = children;
            this[isNodeListSymbol] = true;
        }

        this[internalSymbol] = validateInstance(node, Node);
    }

    /**
     * @private
     * @see https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/yield
     */
    [Symbol.iterator] = function* () {

        /**
         * The end of the generator function is reached. In this case, execution of the generator
         * ends and an IteratorResult is returned to the caller in which the value is undefined and done is true.
         * @see https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/yield
         */
        if (this[internalSymbol] === undefined) {
            return;
        }
        
        // iterator is a nodelist and the main node is only a placeholder
        if (this[isNodeListSymbol] !== true) {
            yield this[internalSymbol];
        }

        if (this[internalSymbol].hasChildNodes()) {
            let childNodes = this[internalSymbol].childNodes;

            for (let node of childNodes) {
                yield* new NodeRecursiveIterator(node);
            }
        }

        return;
    }

    /**
     * @param {function} callback
     * @return {Monster.Types.NodeRecursiveIterator}
     */
    forEach(callback) {
        for (const node of this) {
            callback(node);
        }
        return this;
    }

}


assignToNamespace('Monster.Types', NodeRecursiveIterator);
export {Monster, NodeRecursiveIterator}
