'use strict';

/**
 * @author schukai GmbH
 */
import {assignToNamespace, Monster} from '../namespace.js';
import {Base} from './base.js';
import {isArray, isObject, isPrimitive} from "./is.js";
import {Observer} from "./observer.js";
import {ObserverList} from "./observerlist.js";
import {validateObject} from "./validate.js";
import {extend} from "../data/extend.js";

/**
 * An observer manages a callback function
 *
 * You can call create the class via the monster namespace `new Monster.Types.ProxyObserver()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * new Monster.Types.ProxyObserver()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this class individually.
 *
 * ```
 * <script type="module">
 * import {ProxyObserver} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/proxyobserver.js';
 * new ProxyObserver()
 * </script>
 * ```
 *
 * with the ProxyObserver you can attach observer for observation. with each change at the object to be observed an update takes place.
 *
 * this also applies to nested objects.
 *
 * @example
 *
 * import {ProxyObserver} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/proxyobserver.js';
 * import {Observer} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/observer.js';
 * import {isObject} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/is.js';
 *
 * const o = new Observer(function () { 
 *   if (isObject(this) && this instanceof ProxyObserver) {
 *       // do something (this ist ProxyObserver)
 *       const subject = this.getSubject();
 *       console.log(subject);
 *   }
 * });
 *
 * let realSubject = {
 *   a: {
 *       b: {
 *           c: true
 *       },
 *       d: 9
 *   }
 * }
 *
 * const p = new ProxyObserver(realSubject);
 * p.attachObserver(o);
 * const s = p.getSubject();
 * s.a.b.c = false;
 *
 * @since 1.0.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types
 */
class ProxyObserver extends Base {

    /**
     *
     * @param {object} object
     * @throws {TypeError} value is not a object
     */
    constructor(object) {
        super();

        this.realSubject = validateObject(object);
        this.subject = new Proxy(object, getHandler.call(this));

        this.objectMap = new WeakMap();
        this.objectMap.set(this.realSubject, this.subject);

        this.proxyMap = new WeakMap();
        this.proxyMap.set(this.subject, this.realSubject);

        this.observers = new ObserverList;
    }

    /**
     * get the real object
     *
     * changes to this object are not noticed by the observers, so you can make a large number of changes and inform the observers later.
     *
     * @returns {object}
     */
    getSubject() {
        return this.subject
    }

    /**
     * @since 1.24.0
     * @param {Object} obj
     * @return {Monster.Types.ProxyObserver}
     */
    setSubject(obj) {

        let i, k = Object.keys(this.subject);
        for (i = 0; i < k.length; i++) {
            delete this.subject[k[i]];
        }

        this.subject = extend(this.subject, obj);
        return this;
    }

    /**
     * get the proxied object
     *
     * @returns {object}
     */
    getRealSubject() {
        return this.realSubject
    }

    /**
     * attach a new observer
     *
     * @param {Observer} observer
     * @returns {ProxyObserver}
     */
    attachObserver(observer) {
        this.observers.attach(observer)
        return this;
    }

    /**
     * detach a observer
     *
     * @param {Observer} observer
     * @returns {ProxyObserver}
     */
    detachObserver(observer) {
        this.observers.detach(observer)
        return this;
    }

    /**
     * notify all observer
     *
     * @returns {Promise}
     */
    notifyObservers() {
        return this.observers.notify(this);
    }

    /**
     * @param {Observer} observer
     * @returns {boolean}
     */
    containsObserver(observer) {
        return this.observers.contains(observer)
    }

}

assignToNamespace('Monster.Types', ProxyObserver);
export {Monster, ProxyObserver}

/**
 *
 * @returns {{defineProperty: (function(*=, *=, *=): *), setPrototypeOf: (function(*, *=): boolean), set: (function(*, *, *, *): boolean), get: ((function(*=, *=, *=): (undefined))|*), deleteProperty: ((function(*, *): (boolean))|*)}}
 * @private
 * @see {@link https://gitlab.schukai.com/-/snippets/49}
 */
function getHandler() {

    const proxy = this;

    // https://262.ecma-international.org/9.0/#sec-proxy-object-internal-methods-and-internal-slots
    const handler = {

        // https://262.ecma-international.org/9.0/#sec-proxy-object-internal-methods-and-internal-slots-get-p-receiver
        get: function (target, key, receiver) {

            const value = Reflect.get(target, key, receiver);

            if (typeof key === "symbol") {
                return value;
            }

            if (isPrimitive(value)) {
                return value;
            }

            // set value as proxy if object or array
            if ((isArray(value) || isObject(value))) {
                if (proxy.objectMap.has(value)) {
                    return proxy.objectMap.get(value);
                } else if (proxy.proxyMap.has(value)) {
                    return value;
                } else {
                    let p = new Proxy(value, handler);
                    proxy.objectMap.set(value, p);
                    proxy.proxyMap.set(p, value);
                    return p;
                }

            }

            return value;

        },

        // https://262.ecma-international.org/9.0/#sec-proxy-object-internal-methods-and-internal-slots-set-p-v-receiver 
        set: function (target, key, value, receiver) {

            if (proxy.proxyMap.has(value)) {
                value = proxy.proxyMap.get(value);
            }

            if (proxy.proxyMap.has(target)) {
                target = proxy.proxyMap.get(target);
            }

            let current = Reflect.get(target, key, receiver);
            if (proxy.proxyMap.has(current)) {
                current = proxy.proxyMap.get(current);
            }

            if (current === value) {
                return true;
            }

            let result;
            let descriptor = Reflect.getOwnPropertyDescriptor(target, key);

            if (descriptor === undefined) {
                descriptor = {
                    writable: true,
                    enumerable: true,
                    configurable: true
                }
            }

            descriptor['value'] = value;
            result = Reflect.defineProperty(target, key, descriptor);

            if (typeof key !== "symbol") {
                proxy.observers.notify(proxy);
            }

            return result;
        },


        // https://262.ecma-international.org/9.0/#sec-proxy-object-internal-methods-and-internal-slots-delete-p
        deleteProperty: function (target, key) {
            if (key in target) {
                delete target[key];

                if (typeof key !== "symbol") {
                    proxy.observers.notify(proxy);
                }

                return true;
            }
            return false;
        },

        // https://262.ecma-international.org/9.0/#sec-proxy-object-internal-methods-and-internal-slots-defineownproperty-p-desc
        defineProperty: function (target, key, descriptor) {

            let result = Reflect.defineProperty(target, key, descriptor);
            if (typeof key !== "symbol") {
                proxy.observers.notify(proxy);
            }
            return result;
        },

        // https://262.ecma-international.org/9.0/#sec-proxy-object-internal-methods-and-internal-slots-setprototypeof-v
        setPrototypeOf: function (target, key) {
            let result = Reflect.setPrototypeOf(object1, key);

            if (typeof key !== "symbol") {
                proxy.observers.notify(proxy);
            }

            return result;
        }

    };


    return handler;
}
