'use strict';

/**
 * @author schukai GmbH
 */
import {assignToNamespace, Monster} from '../namespace.js';
import {Base} from './base.js';

/**
 * You can create the instance via the monster namespace `new Monster.Types.Queue()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * new Monster.Types.Queue()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {Queue} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/queue.js';
 * new Queue()
 * </script>
 * ```
 *
 * @example
 *
 * import {Queue} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/queue.js';
 *
 * const queue = new Queue;
 *
 * queue.add(2);
 * queue.add(true);
 * queue.add("Hello");
 * queue.add(4.5);
 *
 * console.log(queue.poll());
 * // ↦ 2
 * console.log(queue.poll());
 * // ↦ true
 * console.log(queue.poll());
 * // ↦ "Hello"
 * console.log(queue.poll());
 * // ↦ 4.5
 * console.log(queue.poll());
 * // ↦ undefined
 *
 *
 * @since 1.4.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types
 * @summary A Queue (Fifo)
 */
class Queue extends Base {

    /**
     *
     */
    constructor() {
        super();
        this.data = [];
    }


    /**
     * @return {boolean}
     */
    isEmpty() {
        return this.data.length === 0;
    }

    /**
     * Read the element at the front of the queue without removing it.
     *
     * @return {*}
     */
    peek() {
        if (this.isEmpty()) {
            return undefined;
        }

        return this.data[0];
    }

    /**
     * Add a new element to the end of the queue.
     *
     * @param {*} value
     * @returns {Queue}
     */
    add(value) {
        this.data.push(value)
        return this;
    }

    /**
     * remove all entries
     *
     * @returns {Queue}
     */
    clear() {
        this.data = [];
        return this;
    }

    /**
     * Remove the element at the front of the queue
     * If the queue is empty, return undefined.
     *
     * @return {*}
     */
    poll() {
        if (this.isEmpty()) {
            return undefined;
        }
        return this.data.shift();
    }


}

assignToNamespace('Monster.Types', Queue);
export {Monster, Queue}
