'use strict';

/**
 * @author schukai GmbH
 */


import {assignToNamespace, Monster} from '../namespace.js';

/**
 * The built-in typeof method is known to have some historical weaknesses. This function tries to provide a better and more accurate result.
 *
 * You can call the method via the monster namespace `Monster.Types.typeOf()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * console.log(Monster.Types.typeOf())
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {typeOf} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/typeof.js';
 * console.log(typeOf())
 * </script>
 * ```
 *
 * @example
 *
 * import {typeOf} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/typeof.js';
 *
 * console.log(typeOf(undefined)); // ↦ undefined
 * console.log(typeOf("")); // ↦ string
 * console.log(typeOf(5)); // ↦ number
 * console.log(typeOf({})); // ↦ object
 * console.log(typeOf([])); // ↦ array
 * console.log(typeOf(new Map)); // ↦ map
 * console.log(typeOf(true)); // ↦ boolean
 *
 * @param {*} value
 * @return {string}
 * @since 1.7.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types
 * @throws {TypeError} value is not a primitive
 */
function typeOf(value) {
    let type = ({}).toString.call(value).match(/\s([a-zA-Z]+)/)[1];
    if ('Object' === type) {
        const results = (/^(class|function)\s+(\w+)/).exec(value.constructor.toString());
        type = (results && results.length > 2) ? results[2] : '';
    }
    return type.toLowerCase();
}

assignToNamespace('Monster.Types', typeOf);
export {
    Monster,
    typeOf
}
