'use strict';

/**
 * @author schukai GmbH
 */

import {assignToNamespace, Monster} from '../namespace.js';
import {Queue} from "./queue.js";
import {validateObject} from "./validate.js";

/**
 * You can call the method via the monster namespace `new Monster.Types.Queue()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * new Monster.Types.UniqueQueue()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {UniqueQueue} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/uniquequeue.js';
 * new UniqueQueue()
 * </script>
 * ```
 *
 * @since 1.4.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types
 * @summary A queue for unique values
 */
class UniqueQueue extends Queue {

    /**
     *
     */
    constructor() {
        super();
        this.unique = new WeakSet();
    }

    /**
     * Add a new element to the end of the queue.
     *
     * @param {object} value
     * @returns {Queue}
     * @throws {TypeError} value is not a object
     */
    add(value) {

        validateObject(value);

        if (!this.unique.has(value)) {
            this.unique.add(value);
            super.add(value);
        }

        return this;
    }

    /**
     * remove all entries
     *
     * @returns {Queue}
     */
    clear() {
        super.clear();
        this.unique = new WeakSet;
        return this;
    }

    /**
     * Remove the element at the front of the queue
     * If the queue is empty, return undefined.
     *
     * @return {object}
     */
    poll() {

        if (this.isEmpty()) {
            return undefined;
        }
        let value = this.data.shift();
        this.unique.delete(value);
        return value;
    }


}

assignToNamespace('Monster.Types', UniqueQueue);
export {Monster, UniqueQueue}
