'use strict';

/**
 * @author schukai GmbH
 */

import {internalSymbol} from "../constants.js";
import {random} from "../math/random.js";
import {assignToNamespace, Monster} from "../namespace.js";
import {isObject} from '../types/is.js';
import {Base} from "./base.js";
import {getGlobalObject} from "./global.js";

/**
 * This is the base class from which all monster classes are derived.
 *
 * You can create an object using the namespace:
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * new Monster.Types.UUID()
 * </script>
 * ```
 *
 * Or directly via the class:
 *
 * ```
 * <script type="module">
 * import {Base} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/uuid.js';
 * new UUID()
 * </script>
 * ```
 *
 * @since 1.25.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types
 * @throws {Error} unsupported
 */
class UUID extends Base {

    /**
     *
     */
    constructor() {
        super();

        let uuid = createWithCrypto();

        if (uuid === undefined) {
            uuid = createWithRandom();
        }


        if (uuid === undefined) {
            throw new Error('unsupported')
        }

        this[internalSymbol] = {
            value: uuid
        }

    }

    /**
     *
     * @return {string}
     */
    toString() {
        return this[internalSymbol]['value'];
    }


}

/**
 * @private
 * @return {string|undefined}
 */
function createWithRandom() {
    return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
        var r = random(0, 65000) * 16 | 0,
            v = ((c === 'x') ? r : (r & 0x3 | 0x8));
        return v.toString(16)[0];
    })
}


/**
 * @private
 * @return {string|undefined}
 */
function createWithCrypto() {
    const crypt = getGlobalObject('crypto');
    if (!isObject(crypt)) return;
    if (typeof crypt?.['randomUUID']) return;
    return crypt.randomUUID();
}


assignToNamespace('Monster.Types', UUID);
export {Monster, UUID}
