'use strict';

/**
 * @author schukai GmbH
 */

import {assignToNamespace, Monster} from '../namespace.js';
import {
    isArray,
    isBoolean,
    isFunction,
    isInstance,
    isInteger,
    isIterable,
    isObject,
    isPrimitive,
    isString,
    isSymbol
} from './is.js';

/**
 * This method checks if the type matches the primitive type. this function is identical to isPrimitive() except that a TypeError is thrown.
 *
 * You can call the method via the monster namespace `Monster.Types.validatePrimitive()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * console.log(Monster.Types.validateIterable('2')) // ↦ TypeError
 * console.log(Monster.Types.validateIterable([])) // ↦ value
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {validateIterable} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/validate.js';
 * console.log(validateIterable('2'))  // ↦ TypeError
 * console.log(validateIterable([]))  // ↦ value
 * </script>
 * ```
 *
 * @param {*} value
 * @return {*}
 * @since 1.2.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types
 * @throws {TypeError} value is not a primitive
 * @see {@link isPrimitive}
 * @see {@link Monster.Types.isPrimitive}
 * @see {@link Monster.Types#isPrimitive}
 */
function validateIterable(value) {
    if (!isIterable(value)) {
        throw new TypeError('value is not iterable')
    }
    return value
}

/**
 * This method checks if the type matches the primitive type. this function is identical to isPrimitive() except that a TypeError is thrown.
 *
 * You can call the method via the monster namespace `Monster.Types.validatePrimitive()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * console.log(Monster.Types.validatePrimitive('2')) // ↦ value
 * console.log(Monster.Types.validatePrimitive([])) // ↦ TypeError
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {validatePrimitive} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/validate.js';
 * console.log(validatePrimitive('2'))  // ↦ value
 * console.log(validatePrimitive([]))  // ↦ TypeError
 * </script>
 * ```
 *
 * @param {*} value
 * @return {*}
 * @since 1.0.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types
 * @throws {TypeError} value is not a primitive
 * @see {@link isPrimitive}
 * @see {@link Monster.Types.isPrimitive}
 * @see {@link Monster.Types#isPrimitive}
 */
function validatePrimitive(value) {
    if (!isPrimitive(value)) {
        throw new TypeError('value is not a primitive')
    }
    return value
}

/**
 * This method checks if the type matches the boolean type. this function is identical to isBoolean() except that a TypeError is thrown.
 *
 * You can call the method via the monster namespace `Monster.Types.validateBoolean()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * console.log(Monster.Types.validateBoolean(true)) // ↦ value
 * console.log(Monster.Types.validateBoolean('2')) // ↦ TypeError
 * console.log(Monster.Types.validateBoolean([])) // ↦ TypeError
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {validateBoolean} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/validate.js';
 * console.log(validateBoolean(false))  // ↦ value
 * console.log(validateBoolean('2'))  // ↦ TypeError
 * console.log(validateBoolean([]))  // ↦ TypeError
 * </script>
 * ```
 *
 * @param {*} value
 * @return {*}
 * @since 1.0.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types

 * @throws {TypeError}  value is not primitive
 */
function validateBoolean(value) {
    if (!isBoolean(value)) {
        throw new TypeError('value is not a boolean')
    }
    return value
}

/**
 * This method checks if the type matches the string type. this function is identical to isString() except that a TypeError is thrown.
 *
 * You can call the method via the monster namespace `Monster.Types.validateString()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * console.log(Monster.Types.validateString('2')) // ↦ value
 * console.log(Monster.Types.validateString([])) // ↦ TypeError
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {validateString} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/validate.js';
 * console.log(validateString('2'))  // ↦ value
 * console.log(validateString([]))  // ↦ TypeError
 * </script>
 * ```
 *
 * @param {*} value
 * @return {*}
 * @since 1.0.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types
 * @throws {TypeError} value is not a string
 */
function validateString(value) {
    if (!isString(value)) {
        throw new TypeError('value is not a string')
    }
    return value
}


/**
 * This method checks if the type matches the object type. this function is identical to isObject() except that a TypeError is thrown.
 *
 * You can call the method via the monster namespace `Monster.Types.validateObject()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * console.log(Monster.Types.validateObject({})) // ↦ value
 * console.log(Monster.Types.validateObject('2')) // ↦ TypeError
 * console.log(Monster.Types.validateObject([])) // ↦ TypeError
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {validateObject} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/validate.js';
 * console.log(validateObject({}))  // ↦ value
 * console.log(validateObject('2'))  // ↦ TypeError
 * console.log(validateObject([]))  // ↦ TypeError
 * </script>
 * ```
 *
 * @param {*} value
 * @return {*}
 * @since 1.0.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types
 * @throws {TypeError} value is not a object
 */
function validateObject(value) {
    if (!isObject(value)) {
        throw new TypeError('value is not a object')
    }
    return value
}

/**
 * This method checks if the type matches the object instance.
 *
 * You can call the method via the monster namespace `Monster.Types.validateInstance()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * console.log(Monster.Types.validateInstance({}, Object)) // ↦ value
 * console.log(Monster.Types.validateInstance('2', Object)) // ↦ TypeError
 * console.log(Monster.Types.validateInstance([], Object)) // ↦ TypeError
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {validateInstance} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/validate.js';
 * console.log(validateInstance({}, Object)) // ↦ value
 * console.log(validateInstance('2', Object)) // ↦ TypeError
 * console.log(validateInstance([], Object)) // ↦ TypeError
 * </script>
 * ```
 *
 * @param {*} value
 * @return {*}
 * @since 1.5.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types
 * @throws {TypeError} value is not an instance of
 */
function validateInstance(value, instance) {
    if (!isInstance(value, instance)) {
        let n = "";
        if (isObject(instance) || isFunction(instance)) {
            n = instance?.['name']
        }

        if (n) {
            n = " " + n;
        }

        throw new TypeError('value is not an instance of' + n)
    }
    return value
}

/**
 * This method checks if the type matches the array type. this function is identical to isArray() except that a TypeError is thrown.
 *
 * You can call the method via the monster namespace `Monster.Types.validateArray()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * console.log(Monster.Types.validateArray('2')) // ↦ TypeError
 * console.log(Monster.Types.validateArray([])) // ↦ value
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {validateArray} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/validate.js';
 * console.log(validateArray('2'))  // ↦ TypeError
 * console.log(validateArray([]))  // ↦ value
 * </script>
 * ```
 *
 * @param {*} value
 * @return {*}
 * @since 1.0.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types
 * @throws {TypeError} value is not an array
 */
function validateArray(value) {
    if (!isArray(value)) {
        throw new TypeError('value is not an array')
    }
    return value
}

/**
 * This method checks if the type matches the symbol type. this function is identical to isSymbol() except that a TypeError is thrown.
 *
 * You can call the method via the monster namespace `Monster.Types.validateSymbol()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * console.log(Monster.Types.validateSymbol('2')) // ↦ TypeError
 * console.log(Monster.Types.validateSymbol([])) // ↦ value
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {validateSymbol} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/validate.js';
 * console.log(validateSymbol('2'))  // ↦ TypeError
 * console.log(validateSymbol())  // ↦ value
 * </script>
 * ```
 *
 * @param {*} value
 * @return {*}
 * @since 1.9.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types
 * @throws {TypeError} value is not an symbol
 */
function validateSymbol(value) {
    if (!isSymbol(value)) {
        throw new TypeError('value is not an symbol')
    }
    return value
}

/**
 * This method checks if the type matches the function type. this function is identical to isFunction() except that a TypeError is thrown.
 *
 * You can call the method via the monster namespace `Monster.Types.validateFunction()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * console.log(Monster.Types.validateFunction(()=>{})) // ↦ value
 * console.log(Monster.Types.validateFunction('2')) // ↦ TypeError
 * console.log(Monster.Types.validateFunction([])) // ↦ TypeError
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {validateFunction} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/validate.js';
 * console.log(validateFunction(()=>{})) // ↦ value
 * console.log(validateFunction('2'))  // ↦ TypeError
 * console.log(validateFunction([]))  // ↦ TypeError
 * </script>
 * ```
 *
 * @param {*} value
 * @return {*}
 * @since 1.0.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types
 * @throws {TypeError} value is not a function
 */
function validateFunction(value) {
    if (!isFunction(value)) {
        throw new TypeError('value is not a function')
    }
    return value
}

/**
 * This method checks if the type is an integer. this function is identical to isInteger() except that a TypeError is thrown.
 *
 * You can call the method via the monster namespace `Monster.Types.validateInteger()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * console.log(Monster.Types.validateInteger(true)) // ↦ TypeError
 * console.log(Monster.Types.validateInteger('2')) // ↦ TypeError
 * console.log(Monster.Types.validateInteger(2)) // ↦ value
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {validateFunction} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/types/validate.js';
 * console.log(validateInteger(true)) // ↦ TypeError
 * console.log(validateInteger('2'))  // ↦ TypeError
 * console.log(validateInteger(2))  // ↦ value
 * </script>
 * ```
 *
 * @param {*} value
 * @return {*}
 * @since 1.4.0
 * @copyright schukai GmbH
 * @memberOf Monster.Types
 * @throws {TypeError} value is not an integer
 */
function validateInteger(value) {
    if (!isInteger(value)) {
        throw new TypeError('value is not an integer')
    }
    return value
}

assignToNamespace('Monster.Types', validatePrimitive, validateBoolean, validateString, validateObject, validateArray, validateFunction, validateIterable, validateInteger);
export {
    Monster,
    validatePrimitive,
    validateBoolean,
    validateString,
    validateObject,
    validateInstance,
    validateArray,
    validateFunction,
    validateIterable,
    validateInteger,
    validateSymbol
}
