'use strict';

/**
 * @author schukai GmbH
 */


import {internalSymbol} from "../constants.js";
import {assignToNamespace, Monster} from '../namespace.js';
import {Base} from "../types/base.js";
import {isInteger} from "../types/is.js";
import {validateFunction, validateInteger} from "../types/validate.js";

/**
 * The dead man's switch allows to set a timer which can be reset again and again within a defined period of time.
 *
 * You can create an instance via the monster namespace `Monster.Util.DeadMansSwitch()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * new Monster.Util.DeadMansSwitch()
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this class individually.
 *
 * ```
 * <script type="module">
 * import {DeadMansSwitch} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/util/deadmansswitch.js';
 * new DeadMansSwitch();
 * </script>
 * ```
 *
 * @example
 * import {DeadMansSwitch} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/util/deadmansswitch.js';
 *
 * const deadmansswitch = new DeadMansSwitch(100, ()=>{
 *   console.log('yeah!')
 *   // ↦ "yeah!"
 * })
 *
 * deadmansswitch.touch(); // from here wait again 100 ms
 * deadmansswitch.touch(200); // from here wait 200 ms
 *
 * @copyright schukai GmbH
 * @since 1.29.0
 * @memberOf Monster.Util
 * @summary Class to be able to execute function chains
 */
class DeadMansSwitch extends Base {

    /**
     * Create new dead man's switch
     *
     * @param {Integer} delay
     * @param {function} callback
     * @throw {TypeError} the arguments must be either integer or functions
     * @throws {TypeError} value is not an integer
     */
    constructor(delay, callback) {
        super();

        init.call(this, validateInteger(delay), validateFunction(callback));
    }

    /**
     *
     * @param {Integer|undefined} [delay]
     */
    touch(delay) {

        if (this[internalSymbol]['isAlreadyRun'] === true) {
            throw new Error('has already run')
        }

        if (isInteger(delay)) {
            this[internalSymbol]['delay'] = delay
        } else if (delay !== undefined) {
            throw new Error('unsupported argument')
        }

        clearTimeout(this[internalSymbol]['timer']);

        initCallback.call(this);

        return this;
    }
}

/**
 * @private
 */
function initCallback() {

    const self = this;

    self[internalSymbol]['timer'] = setTimeout(() => {
        self[internalSymbol]['isAlreadyRun'] = true;
        self[internalSymbol]['callback']();
    }, self[internalSymbol]['delay'])
}

/**
 * @private
 * @param {integer} delay
 * @param {function} callback
 */
function init(delay, callback) {
    const self = this;

    self[internalSymbol] = {
        callback,
        delay,
        isAlreadyRun: false,
        timer: undefined
    };

    initCallback.call(self);

}

assignToNamespace('Monster.Util', DeadMansSwitch);
export {Monster, DeadMansSwitch}
