'use strict';

/**
 * @author schukai GmbH
 */

import {assignToNamespace, Monster} from "../namespace.js";
import {ID} from "../types/id.js";
import {isObject} from "../types/is.js";
import {validateString} from "../types/validate.js";

/**
 * This special trim function allows to trim spaces that have been protected by a special escape character.
 *
 * You can call the method via the monster namespace `Monster.Util.trimSpaces()`.
 *
 * ```
 * <script type="module">
 * import {Monster} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/monster.js';
 * Monster.Util.trimSpaces(" hello ")
 * </script>
 * ```
 *
 * Alternatively, you can also integrate this function individually.
 *
 * ```
 * <script type="module">
 * import {trimSpaces} from 'https://cdn.jsdelivr.net/npm/@schukai/monster@1.30.0/dist/modules/util/trimspaces.js';
 * trimSpaces(' hello \\ ')
 * </script>
 * ```
 * 
 * Hint: One stroke is escaped by the javascript interpreter, the second stroke escapes the stroke.
 * 
 * ```text
 * a\ b  ↦ a b
 * a\\ b ↦ a\ b
 * ```
 * 
 * @since 1.24.0
 * @memberOf Monster.Util
 * @copyright schukai GmbH
 * @param {string} value
 * @return {string}
 * @throws {TypeError} value is not a string
 */
function trimSpaces(value) {

    validateString(value);

    let placeholder = new Map;
    const regex = /((?<pattern>\\(?<char>.)){1})/mig;

    // The separator for args must be escaped
    // undefined string which should not occur normally and is also not a regex
    let result = value.matchAll(regex)

    for (let m of result) {
        let g = m?.['groups'];
        if (!isObject(g)) {
            continue;
        }

        let p = g?.['pattern'];
        let c = g?.['char'];

        if (p && c) {
            let r = '__' + new ID().toString() + '__';
            placeholder.set(r, c);
            value = value.replace(p, r);
        }

    }

    value = value.trim();
    placeholder.forEach((v, k) => {
        value = value.replace(k, '\\' + v)
    })

    return value;

}

assignToNamespace('Monster.Util', trimSpaces);
export {Monster, trimSpaces}
