import {expect} from "chai"
import {Formatter} from "../../../source/text/formatter.js";


describe('Formatter', function () {

    // https://gitlab.schukai.com/oss/libraries/javascript/monster/-/issues/47
    describe('examples', function () {

        it('rfc example should run', function () {
            expect(new Formatter({

                a: {
                    b: {
                        c: "Hello"
                    },
                    d: "World",
                    e: 1
                }

            }).format("${a.b.c} ${a.d | toupper}!")).to.be.equal('Hello WORLD!');
        })

        it('doc example should run', function () {
            expect(new Formatter({

                a: {
                    b: {
                        c: "Hello"
                    },
                    d: "world",
                }

            }).format("${a.b.c} ${a.d | ucfirst}!")).to.be.equal('Hello World!');
        })


    })

    describe('set marker()', function () {


        [
            ['#a#', '#', undefined, 'test'],
            ['{a}', '{', '}', 'test'],
            ['i18n{a}', 'i18n{', '}', 'test'],

        ].forEach(function (data) {

            let a = data.shift()
            let b = data.shift()
            let c = data.shift()
            let d = data.shift()

            it('format ' + a + ' with marker ' + b + ' and ' + c + ' should return  ' + b, function () {

                expect(
                    new Formatter({
                        a: "test"
                    }).setMarker(b, c).format(a)
                ).to.equal(d)
            });
        });


    })

    describe('examples()', function () {

        [
            ['${a | tojson}', "{\"b\":{\"c\":\"Hello\"},\"d\":\"World\",\"e\":1}"],
            ['click ${a.d} times', "click World times"],
            [' ${a.b.c} ', ' Hello '],
            [' ${a.b.c}', ' Hello'],
            ['${a.b.c} ', 'Hello '],
            ['${a.b.c}', 'Hello'],
            ['${a.b.c}${a.b.c}', 'HelloHello'],
            ['${a.b.c} ${a.b.c}', 'Hello Hello'],
            ['${a.b.c} ${a.b.c} ', 'Hello Hello '],
            [' ${a.b.c} ${a.b.c} ', ' Hello Hello '],
            [' ${a.b.c} ${a.d} ', ' Hello World '],
            [' ${a.b.c} ${a.b.c | toupper | length | tostring} ', ' Hello 5 '],

        ].forEach(function (data) {

            let a = data.shift()
            let b = data.shift()

            it('format ' + a + ' should return ' + b, function () {

                let obj = {
                    a: {
                        b: {
                            c: "Hello"
                        },
                        d: "World",
                        e: 1
                    }
                }


                expect(new Formatter(obj).format(a)).is.equal(b)
            });
        });

    });

    describe('Marker in marker', function () {

        let text = '${mykey${subkey}}';
        let expected = '1';

        it('format ' + text + ' should ' + expected, function () {

            let obj = {
                mykey2: "1",
                subkey: "2"
            };

            expect(new Formatter(obj).format(text)).is.equal(expected)

        });


    });

    describe('setParameterChars()', function () {

        it('setParameterChars() should return Instance', function () {
            expect(new Formatter({}).setParameterChars('a', 'b')).is.instanceof(Formatter);

        });


    });


    describe('with callbacks', function () {

        it('add callback', function () {
            const formatter = new Formatter({
                x: '1'
            }, {
                callbacks: {
                    my: (value) => {
                        return "!" + value + "!"
                    }
                }
            });

            expect(formatter.format('${x | call:my}')).is.equal('!1!');

        });


    });

    describe('Marker in marker with parameter', function () {

        let text = '${mykey::mykey=${subkey}}';
        let expected = '2';

        it('format ' + text + ' should ' + expected, function () {

            let obj = {
                subkey: "2"
            };

            expect(new Formatter(obj).format(text)).is.equal(expected)

        });
    });

    describe('exceptions', function () {

        [
            ['${a.b.x}', TypeError],
            ['${a.b.d | toupper | length}', TypeError],
            ['${a.b.d}', TypeError],  // a.b.d return undefined by pathfinder 
            ['${a.b.d | tolower}', TypeError],  // a.b.d return undefined by pathfinder 
            ['${a | }', Error],

        ].forEach(function (data) {

            let a = data.shift()
            let b = data.shift()

            it('format ' + a + ' should throw ' + typeof b, function () {

                expect(() => {
                        new Formatter({
                            a: {
                                b: {
                                    c: "test",
                                    d: 4
                                }
                            }
                        }).format(a)
                    }
                ).to.throw(b);
            });
        });

    });


});